import pytest
import pickle

from collectionish._sentry import Sentry

SENTRY = Sentry()


def test_sentry_may_not_be_subclassed():
    with pytest.raises(TypeError):

        class SubSentry(Sentry):

            pass


def test_setattr_on_sentry_fails(sentry: Sentry = SENTRY):
    with pytest.raises(AttributeError):
        sentry.apples = 1


def test_sentry_evaluates_false(sentry: Sentry = SENTRY):
    if sentry:
        pytest.fail('Sentry objects should eval False')


def test_pickle_sentry(subtests):
    unpickled = pickle.loads(pickle.dumps(SENTRY))

    with subtests.test(msg='test equality on unpickled sentry'):
        assert SENTRY == unpickled

    with subtests.test(msg='test_sentry_evaluates_false(unpickled)'):
        test_sentry_evaluates_false(unpickled)

    with subtests.test(msg='test_setattr_on_sentry_fails(unpickled)', i=3):
        test_setattr_on_sentry_fails(unpickled)



from typing import Union, Dict
from collections import ChainMap


class AncestorChainMap(ChainMap):

    """ChainMap variant that allows for the insertion of parents and ancestors.

    Example:
        >>> from collectionish import AncestorChainMap
        >>>
        >>> child = AncestorChainMap({'b': 3, 'c': 3})
        >>> child
        AncestorChainMap({'b': 3, 'c': 3})

        add oldest ancestor:

        >>> child.add_ancestor({'a': 1, 'b': 1})
        >>> child
        AncestorChainMap({'b': 3, 'c': 3}, {'a': 1, 'b': 1})

        add direct parent:

        >>> child.add_parent({'a': 2, 'b': 2})
        >>> child
        AncestorChainMap({'b': 3, 'c': 3}, {'a': 2, 'b': 2}, {'a': 1, 'b': 1})

    """

    def add_parent(self, parent: Union[Dict, ChainMap]):
        """add a parent to the current chain map"""
        self.maps.insert(1, dict(parent))

    def add_ancestor(self, ancestor: Union[Dict, ChainMap]):
        """add a oldest ancestor to the current chain map"""
        return self.maps.append(dict(ancestor))



from typing import Any, ClassVar, Optional


class Sentry:

    """A Basic falsey singleton type useful for when ``None`` actually means something.

    Only one instance of Sentry may be created (within a session).
    To prevent weirdness in with import reloads or multiprocessing we test equality /
    based on is tested based on class membership and disallow subclassing.

    The following actions are is forbidden on the Sentry type:
        - subclassing
        - setting attributes
    """

    __slots__ = ()
    _instance: ClassVar[Optional['Sentry']] = None

    def __init_subclass__(cls, *args, **kwargs):
        raise TypeError('Sentry may not be subclassed')

    def __new__(cls) -> 'Sentry':
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Sentry)

    def __bool__(self) -> bool:
        return False

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}()'



import pytest
from collections import ChainMap
from collectionish import AncestorChainMap


@pytest.mark.parametrize(
    'parent,', [{'a': 1, 'b': 2}, ChainMap({'a': 1, 'b': 2}), AncestorChainMap({'a': 1, 'b': 2})]
)
def test_add_parent(parent):
    child = AncestorChainMap({'b': 3, 'c': 4})
    child.add_parent(parent)
    assert dict(child) == {'a': 1, 'b': 3, 'c': 4}


@pytest.mark.parametrize(
    'ancestor,', [{'a': 1, 'd': 1}, ChainMap({'a': 1, 'd': 1}), AncestorChainMap({'a': 1, 'd': 1})]
)
def test_add_ancestor(ancestor):
    child = AncestorChainMap({'b': 3, 'c': 3}, {'a': 2, 'b': 2})
    child.add_ancestor(ancestor)
    assert dict(child) == {'a': 2, 'b': 3, 'c': 3, 'd': 1}
